@@ -1,6 +1,8 @@
 import math
+import logging
 from random import random
 from functools import partial
+from pathlib import Path
 
 import torch
 from torch import nn, Tensor, einsum, IntTensor, FloatTensor, BoolTensor
@@ -8,26 +10,30 @@ from torch.nn import Module
 import torch.nn.functional as F
 
 import torchode as to
-
 from torchdiffeq import odeint
 
 from beartype import beartype
-from beartype.typing import Tuple, Optional
+from beartype.typing import Tuple, Optional, List
 
 from einops.layers.torch import Rearrange
 from einops import rearrange, repeat, reduce, pack, unpack
 
 from voicebox_pytorch.attend import Attend
 
-from naturalspeech2_pytorch.aligner import Aligner, ForwardSumLoss, maximum_path
+from naturalspeech2_pytorch.aligner import Aligner, ForwardSumLoss, BinLoss, maximum_path
+from naturalspeech2_pytorch.utils.tokenizer import Tokenizer
+from naturalspeech2_pytorch.naturalspeech2_pytorch import generate_mask_from_repeats
 
 from audiolm_pytorch import EncodecWrapper
+from spear_tts_pytorch import TextToSemantic
 
 import torchaudio.transforms as T
-from torchaudio.functional import DB_to_amplitude
+from torchaudio.functional import DB_to_amplitude, resample
 
 from vocos import Vocos
 
+LOGGER = logging.getLogger(__file__)
+
 # helper functions
 
 def exists(val):
@@ -64,6 +70,49 @@ def prob_mask_like(shape, prob, device):
     else:
         return torch.zeros(shape, device = device).float().uniform_(0, 1) < prob
 
+def reduce_masks_with_and(*masks):
+    masks = [*filter(exists, masks)]
+
+    if len(masks) == 0:
+        return None
+
+    mask, *rest_masks = masks
+
+    for rest_mask in rest_masks:
+        mask = mask & rest_mask
+
+    return mask
+
+def interpolate_1d(t, length, mode = 'bilinear'):
+    " pytorch does not offer interpolation 1d, so hack by converting to 2d "
+
+    dtype = t.dtype
+    t = t.float()
+
+    implicit_one_channel = t.ndim == 2
+    if implicit_one_channel:
+        t = rearrange(t, 'b n -> b 1 n')
+
+    t = rearrange(t, 'b d n -> b d n 1')
+    t = F.interpolate(t, (length, 1), mode = mode)
+    t = rearrange(t, 'b d n 1 -> b d n')
+
+    if implicit_one_channel:
+        t = rearrange(t, 'b 1 n -> b n')
+
+    t = t.to(dtype)
+    return t
+
+def curtail_or_pad(t, target_length):
+    length = t.shape[-2]
+
+    if length > target_length:
+        t = t[..., :target_length, :]
+    elif length < target_length:
+        t = F.pad(t, (0, 0, 0, target_length - length), value = 0.)
+
+    return t
+
 # mask construction helpers
 
 def mask_from_start_end_indices(
@@ -215,6 +264,7 @@ class Attention(Module):
         dim,
         dim_head = 64,
         heads = 8,
+        dropout=0,
         flash = False
     ):
         super().__init__()
@@ -222,7 +272,7 @@ class Attention(Module):
         self.scale = dim_head ** -0.5
         dim_inner = dim_head * heads
 
-        self.attend = Attend(flash = flash)
+        self.attend = Attend(dropout, flash = flash)
         self.to_qkv = nn.Linear(dim, dim_inner * 3, bias = False)
         self.to_out = nn.Linear(dim_inner, dim, bias = False)
 
@@ -260,9 +310,12 @@ class Transformer(Module):
         dim_head = 64,
         heads = 8,
         ff_mult = 4,
+        attn_dropout=0,
         attn_flash = False,
         adaptive_rmsnorm = False,
-        adaptive_rmsnorm_cond_dim_in = None
+        adaptive_rmsnorm_cond_dim_in = None,
+        use_unet_skip_connection = False,
+        skip_connect_scale = None
     ):
         super().__init__()
         assert divisible_by(depth, 2)
@@ -275,14 +328,16 @@ class Transformer(Module):
         else:
             rmsnorm_klass = RMSNorm
 
+        self.skip_connect_scale = default(skip_connect_scale, 2 ** -0.5)
+
         for ind in range(depth):
             layer = ind + 1
-            has_skip = layer > (depth // 2)
+            has_skip = use_unet_skip_connection and layer > (depth // 2)
 
             self.layers.append(nn.ModuleList([
                 nn.Linear(dim * 2, dim) if has_skip else None,
                 rmsnorm_klass(dim = dim),
-                Attention(dim = dim, dim_head = dim_head, heads = heads, flash = attn_flash),
+                Attention(dim=dim, dim_head=dim_head, heads=heads, dropout=attn_dropout, flash=attn_flash),
                 rmsnorm_klass(dim = dim),
                 FeedForward(dim = dim, mult = ff_mult)
             ]))
@@ -292,6 +347,7 @@ class Transformer(Module):
     def forward(
         self,
         x,
+        mask = None,
         adaptive_rmsnorm_cond = None
     ):
         skip_connects = []
@@ -310,11 +366,12 @@ class Transformer(Module):
             if not exists(skip_combiner):
                 skip_connects.append(x)
             else:
-                x = torch.cat((x, skip_connects.pop()), dim = -1)
+                skip_connect = skip_connects.pop() * self.skip_connect_scale
+                x = torch.cat((x, skip_connect), dim = -1)
                 x = skip_combiner(x)
 
             attn_input = attn_prenorm(x, **rmsnorm_kwargs)
-            x = attn(attn_input, rotary_emb = rotary_emb) + x
+            x = attn(attn_input, mask = mask, rotary_emb = rotary_emb) + x
 
             ff_input = ff_prenorm(x, **rmsnorm_kwargs) 
             x = ff(ff_input) + x
@@ -350,6 +407,10 @@ class MelVoco(AudioEncoderDecoder):
 
         self.vocos = Vocos.from_pretrained(pretrained_vocos_path)
 
+    @property
+    def downsample_factor(self):
+        raise NotImplementedError
+
     @property
     def latent_dim(self):
         return self.num_mels
@@ -391,15 +452,21 @@ class EncodecVoco(AudioEncoderDecoder):
     def __init__(
         self,
         *,
+        sampling_rate = 24000,
         pretrained_vocos_path = 'charactr/vocos-encodec-24khz',
         bandwidth_id = 2
     ):
         super().__init__()
+        self.sampling_rate = sampling_rate
         self.encodec = EncodecWrapper()
         self.vocos = Vocos.from_pretrained(pretrained_vocos_path)
 
         self.register_buffer('bandwidth_id', torch.tensor([bandwidth_id]))
 
+    @property
+    def downsample_factor(self):
+        return self.encodec.downsample_factor
+
     @property
     def latent_dim(self):
         return self.encodec.codebook_dim
@@ -423,11 +490,13 @@ class EncodecVoco(AudioEncoderDecoder):
 # both duration and main denoising model are transformers
 
 class DurationPredictor(Module):
+    @beartype
     def __init__(
         self,
         *,
-        num_phoneme_tokens,
         audio_enc_dec: Optional[AudioEncoderDecoder] = None,
+        tokenizer: Optional[Tokenizer] = None,
+        num_phoneme_tokens: Optional[int] = None,
         dim_phoneme_emb = 512,
         dim = 512,
         depth = 10,
@@ -436,6 +505,7 @@ class DurationPredictor(Module):
         ff_mult = 4,
         conv_pos_embed_kernel_size = 31,
         conv_pos_embed_groups = None,
+        attn_dropout=0,
         attn_flash = False,
         p_drop_prob = 0.2, # p_drop in paper
         frac_lengths_mask: Tuple[float, float] = (0.1, 1.),
@@ -443,6 +513,8 @@ class DurationPredictor(Module):
     ):
         super().__init__()
 
+        # audio encoder / decoder
+
         self.audio_enc_dec = audio_enc_dec
 
         if exists(audio_enc_dec) and dim != audio_enc_dec.latent_dim:
@@ -450,13 +522,24 @@ class DurationPredictor(Module):
         else:
             self.proj_in = nn.Identity()
 
-        self.null_phoneme_id = num_phoneme_tokens # use last phoneme token as null token for CFG
-        self.to_phoneme_emb = nn.Embedding(num_phoneme_tokens + 1, dim_phoneme_emb)
+        # phoneme related
+
+        assert not (exists(tokenizer) and exists(num_phoneme_tokens)), 'if a phoneme tokenizer was passed into duration module, number of phoneme tokens does not need to be specified'
+
+        if not exists(tokenizer) and not exists(num_phoneme_tokens):
+            tokenizer = Tokenizer() # default to english phonemes with espeak
+
+        if exists(tokenizer):
+            num_phoneme_tokens = tokenizer.vocab_size
+
+        self.tokenizer = tokenizer
+
+        self.to_phoneme_emb = nn.Embedding(num_phoneme_tokens, dim_phoneme_emb)
 
         self.p_drop_prob = p_drop_prob
         self.frac_lengths_mask = frac_lengths_mask
 
-        self.to_embed = nn.Linear(dim * 2 + dim_phoneme_emb, dim)
+        self.to_embed = nn.Linear(dim + dim_phoneme_emb, dim)
 
         self.null_cond = nn.Parameter(torch.zeros(dim))
 
@@ -472,6 +555,7 @@ class DurationPredictor(Module):
             dim_head = dim_head,
             heads = heads,
             ff_mult = ff_mult,
+            attn_dropout=attn_dropout,
             attn_flash = attn_flash
         )
 
@@ -492,20 +576,45 @@ class DurationPredictor(Module):
     def device(self):
         return next(self.parameters()).device
 
+    def align_phoneme_ids_with_durations(self, phoneme_ids, durations):
+        repeat_mask = generate_mask_from_repeats(durations.clamp(min = 1))
+        aligned_phoneme_ids = einsum('b i, b i j -> b j', phoneme_ids.float(), repeat_mask.float()).long()
+        return aligned_phoneme_ids
+
     @torch.inference_mode()
+    @beartype
     def forward_with_cond_scale(
         self,
         *args,
+        texts: Optional[List[str]] = None,
+        phoneme_ids = None,
         cond_scale = 1.,
+        return_aligned_phoneme_ids = False,
         **kwargs
     ):
-        logits = self.forward(*args, cond_drop_prob = 0., **kwargs)
+        if exists(texts):
+            phoneme_ids = self.tokenizer.texts_to_tensor_ids(texts)
+
+        forward_kwargs = dict(
+            return_aligned_phoneme_ids = False,
+            phoneme_ids = phoneme_ids
+        )
+
+        durations = self.forward(*args, cond_drop_prob = 0., **forward_kwargs, **kwargs)
 
         if cond_scale == 1.:
-            return logits
+            if not return_aligned_phoneme_ids:
+                return durations
 
-        null_logits = self.forward(*args, cond_drop_prob = 1., **kwargs)
-        return null_logits + (logits - null_logits) * cond_scale
+            return durations, self.align_phoneme_ids_with_durations(phoneme_ids, durations)
+
+        null_durations = self.forward(*args, cond_drop_prob = 1., **forward_kwargs, **kwargs)
+        scaled_durations = null_durations + (durations - null_durations) * cond_scale
+
+        if not return_aligned_phoneme_ids:
+            return scaled_durations
+
+        return scaled_durations, self.align_phoneme_ids_with_durations(phoneme_ids, scaled_durations)
 
     @beartype
     def forward_aligner(
@@ -534,36 +643,44 @@ class DurationPredictor(Module):
         alignment_soft = rearrange(alignment_soft, 'b 1 t1 t2 -> b t2 t1')
         return alignment_hard, alignment_soft, alignment_logprob, alignment_mas
 
+    @beartype
     def forward(
         self,
-        x,
         *,
-        phoneme_ids,
-        mel,
         cond,
+        texts: Optional[List[str]] = None,
+        phoneme_ids = None,
         cond_drop_prob = 0.,
         target = None,
-        mask = None,
+        cond_mask = None,
+        mel = None,
         phoneme_len = None,
         mel_len = None,
         phoneme_mask = None,
         mel_mask = None,
+        self_attn_mask = None,
+        return_aligned_phoneme_ids = False
     ):
         batch, seq_len, cond_dim = cond.shape
-        assert cond_dim == x.shape[-1]
 
-        x, cond = map(self.proj_in, (x, cond))
+        cond = self.proj_in(cond)
+
+        # text to phonemes, if tokenizer is given
+
+        if not exists(phoneme_ids):
+            assert exists(self.tokenizer)
+            phoneme_ids = self.tokenizer.texts_to_tensor_ids(texts)
 
         # construct mask if not given
 
-        if not exists(mask):
+        if not exists(cond_mask):
             if coin_flip():
                 frac_lengths = torch.zeros((batch,), device = self.device).float().uniform_(*self.frac_lengths_mask)
-                mask = mask_from_frac_lengths(seq_len, frac_lengths)
+                cond_mask = mask_from_frac_lengths(seq_len, frac_lengths)
             else:
-                mask = prob_mask_like((batch, seq_len), self.p_drop_prob, self.device)
+                cond_mask = prob_mask_like((batch, seq_len), self.p_drop_prob, self.device)
 
-        cond = cond * rearrange(~mask, '... -> ... 1')
+        cond = cond * rearrange(~cond_mask, '... -> ... 1')
 
         # classifier free guidance
 
@@ -576,47 +693,65 @@ class DurationPredictor(Module):
                 cond
             )
 
-            phoneme_ids = torch.where(
-                rearrange(cond_drop_mask, '... -> ... 1'),
-                self.null_phoneme_id,
-                phoneme_ids
-            )
+        # phoneme id of -1 is padding
+
+        if not exists(self_attn_mask):
+            self_attn_mask = phoneme_ids != -1
+
+        phoneme_ids = phoneme_ids.clamp(min = 0)
+
+        # get phoneme embeddings
 
         phoneme_emb = self.to_phoneme_emb(phoneme_ids)
 
+        # force condition to be same length as input phonemes
+
+        cond = curtail_or_pad(cond, phoneme_ids.shape[-1])
+
+        # combine audio, phoneme, conditioning
+
+        embed = torch.cat((phoneme_emb, cond), dim = -1)
+        x = self.to_embed(embed)
+
+        x = self.conv_embed(x) + x
+
+        x = self.transformer(
+            x,
+            mask = self_attn_mask
+        )
+
+        durations = self.to_pred(x)
+
+        if not self.training:
+            if not return_aligned_phoneme_ids:
+                return durations
+
+            return durations, self.align_phoneme_ids_with_durations(phoneme_ids, durations)
+
         # aligner
         # use alignment_hard to oversample phonemes
         # Duration Predictor should predict the duration of unmasked phonemes where target is masked alignment_hard
 
-        should_align = all([exists(el) for el in (phoneme_len, mel_len, phoneme_mask, mel_mask)])
+        assert all([exists(el) for el in (phoneme_len, mel_len, phoneme_mask, mel_mask)]), 'need to pass phoneme_len, mel_len, phoneme_mask, mel_mask, to train duration predictor module'
 
-        if should_align:
-            alignment_hard, _, alignment_logprob, _ = self.forward_aligner(phoneme_emb, phoneme_mask, mel, mel_mask)
-            target = alignment_hard
-        # combine audio, phoneme, conditioning
+        alignment_hard, _, alignment_logprob, _ = self.forward_aligner(phoneme_emb, phoneme_mask, mel, mel_mask)
+        target = alignment_hard
 
-        embed = torch.cat((x, phoneme_emb, cond), dim = -1)
-        x = self.to_embed(embed)
-
-        x = self.conv_embed(x) + x
-
-        x = self.transformer(x)
-
-        x = self.to_pred(x)
-
-        if not exists(target):
-            return x
+        if exists(self_attn_mask):
+            loss_mask = cond_mask & self_attn_mask
+        else:
+            loss_mask = self_attn_mask
 
         if not exists(mask):
             return F.l1_loss(x, target)
 
         loss = F.l1_loss(x, target, reduction = 'none')
-        loss = loss.masked_fill(~mask, 0.)
+        loss = loss.masked_fill(~loss_mask, 0.)
 
         # masked mean
 
         num = reduce(loss, 'b n -> b', 'sum')
-        den = mask.sum(dim = -1).clamp(min = 1e-5)
+        den = loss_mask.sum(dim = -1).clamp(min = 1e-5)
         loss = num / den
         loss = loss.mean()
         
@@ -634,10 +769,10 @@ class VoiceBox(Module):
     def __init__(
         self,
         *,
-        num_phoneme_tokens,
+        num_cond_tokens = None,
         audio_enc_dec: Optional[AudioEncoderDecoder] = None,
         dim_in = None,
-        dim_phoneme_emb = 1024,
+        dim_cond_emb = 1024,
         dim = 1024,
         depth = 24,
         dim_head = 64,
@@ -646,9 +781,11 @@ class VoiceBox(Module):
         time_hidden_dim = None,
         conv_pos_embed_kernel_size = 31,
         conv_pos_embed_groups = None,
+        attn_dropout=0,
         attn_flash = False,
         p_drop_prob = 0.3, # p_drop in paper
-        frac_lengths_mask: Tuple[float, float] = (0.7, 1.)
+        frac_lengths_mask: Tuple[float, float] = (0.7, 1.),
+        condition_on_text = True
     ):
         super().__init__()
         dim_in = default(dim_in, dim)
@@ -668,13 +805,23 @@ class VoiceBox(Module):
             nn.SiLU()
         )
 
-        self.null_phoneme_id = num_phoneme_tokens # use last phoneme token as null token for CFG
-        self.to_phoneme_emb = nn.Embedding(num_phoneme_tokens + 1, dim_phoneme_emb)
+        assert not (condition_on_text and not exists(num_cond_tokens)), 'number of conditioning tokens must be specified (whether phonemes or semantic token ids) if training conditional voicebox'
+
+        if not condition_on_text:
+            dim_cond_emb = 0
+
+        self.dim_cond_emb = dim_cond_emb
+        self.condition_on_text = condition_on_text
+        self.num_cond_tokens = num_cond_tokens
+
+        if condition_on_text:
+            self.null_cond_id = num_cond_tokens # use last phoneme token as null token for CFG
+            self.to_cond_emb = nn.Embedding(num_cond_tokens + 1, dim_cond_emb)
 
         self.p_drop_prob = p_drop_prob
         self.frac_lengths_mask = frac_lengths_mask
 
-        self.to_embed = nn.Linear(dim_in * 2 + dim_phoneme_emb, dim)
+        self.to_embed = nn.Linear(dim_in * 2 + dim_cond_emb, dim)
 
         self.null_cond = nn.Parameter(torch.zeros(dim_in))
 
@@ -690,6 +837,7 @@ class VoiceBox(Module):
             dim_head = dim_head,
             heads = heads,
             ff_mult = ff_mult,
+            attn_dropout=attn_dropout,
             attn_flash = attn_flash,
             adaptive_rmsnorm = True,
             adaptive_rmsnorm_cond_dim_in = time_hidden_dim
@@ -722,19 +870,27 @@ class VoiceBox(Module):
         self,
         x,
         *,
-        phoneme_ids,
-        cond,
         times,
+        cond_token_ids,
+        self_attn_mask = None,
         cond_drop_prob = 0.1,
         target = None,
-        mask = None,
+        cond = None,
+        cond_mask = None
     ):
-        batch, seq_len, cond_dim = cond.shape
-        assert cond_dim == x.shape[-1]
-
         # project in, in case codebook dim is not equal to model dimensions
 
-        x, cond = map(self.proj_in, (x, cond))
+        x = self.proj_in(x)
+
+        if exists(cond):
+            cond = self.proj_in(cond)
+
+        cond = default(cond, x)
+
+        # shapes
+
+        batch, seq_len, cond_dim = cond.shape
+        assert cond_dim == x.shape[-1]
 
         # auto manage shape of times, for odeint times
 
@@ -744,16 +900,25 @@ class VoiceBox(Module):
         if times.ndim == 1 and times.shape[0] == 1:
             times = repeat(times, '1 -> b', b = cond.shape[0])
 
-        # construct mask if not given
+        # construct conditioning mask if not given
 
-        if not exists(mask):
-            if coin_flip():
-                frac_lengths = torch.zeros((batch,), device = self.device).float().uniform_(*self.frac_lengths_mask)
-                mask = mask_from_frac_lengths(seq_len, frac_lengths)
-            else:
-                mask = prob_mask_like((batch, seq_len), self.p_drop_prob, self.device)
+        if self.training:
+            if not exists(cond_mask):
+                if coin_flip():
+                    frac_lengths = torch.zeros((batch,), device = self.device).float().uniform_(*self.frac_lengths_mask)
+                    cond_mask = mask_from_frac_lengths(seq_len, frac_lengths)
+                else:
+                    cond_mask = prob_mask_like((batch, seq_len), self.p_drop_prob, self.device)
+        else:
+            if not exists(cond_mask):
+                cond_mask = torch.ones((batch, seq_len), device = cond.device, dtype = torch.bool)
 
-        cond = cond * rearrange(~mask, '... -> ... 1')
+        cond_mask_with_pad_dim = rearrange(cond_mask, '... -> ... 1')
+
+        # as described in section 3.2
+
+        x = x * cond_mask_with_pad_dim
+        cond = cond * ~cond_mask_with_pad_dim
 
         # classifier free guidance
 
@@ -766,14 +931,34 @@ class VoiceBox(Module):
                 cond
             )
 
-            phoneme_ids = torch.where(
+            cond_ids = torch.where(
                 rearrange(cond_drop_mask, '... -> ... 1'),
-                self.null_phoneme_id,
-                phoneme_ids
+                self.null_cond_id,
+                cond_token_ids
             )
 
-        phoneme_emb = self.to_phoneme_emb(phoneme_ids)
-        embed = torch.cat((x, phoneme_emb, cond), dim = -1)
+        # phoneme or semantic conditioning embedding
+
+        cond_emb = None
+
+        if self.condition_on_text:
+            cond_emb = self.to_cond_emb(cond_token_ids)
+
+            cond_emb_length = cond_emb.shape[-2]
+            if cond_emb_length != seq_len:
+                cond_emb = rearrange(cond_emb, 'b n d -> b d n')
+                cond_emb = interpolate_1d(cond_emb, seq_len)
+                cond_emb = rearrange(cond_emb, 'b d n -> b n d')
+
+                if exists(self_attn_mask):
+                    self_attn_mask = interpolate_1d(self_attn_mask, seq_len)
+
+        # concat source signal, semantic / phoneme conditioning embed, and conditioning
+        # and project
+
+        to_concat = [*filter(exists, (x, cond_emb, cond))]
+        embed = torch.cat(to_concat, dim = -1)
+
         x = self.to_embed(embed)
 
         x = self.conv_embed(x) + x
@@ -782,7 +967,11 @@ class VoiceBox(Module):
 
         # attend
 
-        x = self.transformer(x, adaptive_rmsnorm_cond = time_emb)
+        x = self.transformer(
+            x,
+            mask = self_attn_mask,
+            adaptive_rmsnorm_cond = time_emb
+        )
 
         x = self.to_pred(x)
 
@@ -791,18 +980,20 @@ class VoiceBox(Module):
         if not exists(target):
             return x
 
-        if not exists(mask):
+        loss_mask = reduce_masks_with_and(cond_mask, self_attn_mask)
+
+        if not exists(loss_mask):
             return F.mse_loss(x, target)
 
         loss = F.mse_loss(x, target, reduction = 'none')
 
         loss = reduce(loss, 'b n d -> b n', 'mean')
-        loss = loss.masked_fill(~mask, 0.)
+        loss = loss.masked_fill(~loss_mask, 0.)
 
         # masked mean
 
         num = reduce(loss, 'b n -> b', 'sum')
-        den = mask.sum(dim = -1).clamp(min = 1e-5)
+        den = loss_mask.sum(dim = -1).clamp(min = 1e-5)
         loss = num / den
 
         return loss.mean()
@@ -810,13 +1001,15 @@ class VoiceBox(Module):
 # wrapper for the CNF
 
 def is_probably_audio_from_shape(t):
-    return t.ndim == 2 or (t.ndim == 3 and t.shape[1] == 1)
+    return exists(t) and (t.ndim == 2 or (t.ndim == 3 and t.shape[1] == 1))
 
 class ConditionalFlowMatcherWrapper(Module):
     @beartype
     def __init__(
         self,
         voicebox: VoiceBox,
+        text_to_semantic: Optional[TextToSemantic] = None,
+        duration_predictor: Optional[DurationPredictor] = None,
         sigma = 0.,
         ode_atol = 1e-5,
         ode_rtol = 1e-5,
@@ -830,6 +1023,15 @@ class ConditionalFlowMatcherWrapper(Module):
         self.sigma = sigma
 
         self.voicebox = voicebox
+        self.condition_on_text = voicebox.condition_on_text
+
+        assert not (not self.condition_on_text and exists(text_to_semantic)), 'TextToSemantic should not be passed in if not conditioning on text'
+        assert not (exists(text_to_semantic) and not exists(text_to_semantic.wav2vec)), 'the wav2vec module must exist on the TextToSemantic, if being used to condition on text'
+
+        self.text_to_semantic = text_to_semantic
+        self.duration_predictor = duration_predictor
+
+        assert exists(text_to_semantic) ^ exists(duration_predictor), 'you should use either TextToSemantic from Spear-TTS, or DurationPredictor for the text / phoneme to audio alignment, but not both'
 
         self.cond_drop_prob = cond_drop_prob
 
@@ -847,20 +1049,31 @@ class ConditionalFlowMatcherWrapper(Module):
     def device(self):
         return next(self.parameters()).device
 
+    def load(self, path, strict = True):
+        # return pkg so the trainer can access it
+        path = Path(path)
+        assert path.exists()
+        pkg = torch.load(str(path), map_location = 'cpu')
+        self.load_state_dict(pkg['model'], strict = strict)
+        return pkg
+
     @torch.inference_mode()
     def sample(
         self,
         *,
-        phoneme_ids,
-        cond,
-        mask = None,
+        cond = None,
+        texts: Optional[List[str]] = None,
+        text_token_ids: Optional[Tensor] = None,
+        semantic_token_ids: Optional[Tensor] = None,
+        phoneme_ids: Optional[Tensor] = None,
+        cond_mask = None,
         steps = 3,
         cond_scale = 1.,
-        decode_to_audio = True
+        decode_to_audio = True,
+        max_semantic_token_ids = 2048,
+        spec_decode = False,
+        spec_decode_gamma = 5 # could be higher, since speech is probably easier than text, needs to be tested
     ):
-        shape = cond.shape
-        batch = shape[0]
-
         # take care of condition as raw audio
 
         cond_is_raw_audio = is_probably_audio_from_shape(cond)
@@ -871,6 +1084,73 @@ class ConditionalFlowMatcherWrapper(Module):
             self.voicebox.audio_enc_dec.eval()
             cond = self.voicebox.audio_enc_dec.encode(cond)
 
+        # setup text conditioning, either coming from duration model (as phoneme ids)
+        # for coming from text-to-semantic module from spear-tts paper, as (semantic ids)
+
+        num_cond_inputs = sum([*map(exists, (texts, text_token_ids, semantic_token_ids, phoneme_ids))])
+        assert num_cond_inputs <= 1
+
+        self_attn_mask = None
+        cond_token_ids = None
+
+        if self.condition_on_text:
+            if exists(self.text_to_semantic) or exists(semantic_token_ids):
+                assert not exists(phoneme_ids)
+
+                if not exists(semantic_token_ids):
+                    self.text_to_semantic.eval()
+
+                    semantic_token_ids, self_attn_mask = self.text_to_semantic.generate(
+                        source = default(text_token_ids, texts),
+                        source_type = 'text',
+                        target_type = 'speech',
+                        max_length = max_semantic_token_ids,
+                        return_target_mask = True,
+                        spec_decode = spec_decode,
+                        spec_decode_gamma = spec_decode_gamma
+                    )
+
+                cond_token_ids = semantic_token_ids
+
+            elif exists(self.duration_predictor):
+                self.duration_predictor.eval()
+
+                durations, aligned_phoneme_ids = self.duration_predictor.forward_with_cond_scale(
+                    cond = cond,
+                    texts = texts,
+                    phoneme_ids = phoneme_ids,
+                    return_aligned_phoneme_ids = True
+                )
+
+                cond_token_ids = aligned_phoneme_ids
+
+            cond_tokens_seq_len = cond_token_ids.shape[-1]
+
+            if exists(cond):
+                if exists(self.text_to_semantic):
+                    # calculate the correct conditioning length for text to semantic
+                    # based on the sampling freqs of wav2vec and audio-enc-dec, as well as downsample factor
+                    # (cond_time x cond_sampling_freq / cond_downsample_factor) == (audio_time x audio_sampling_freq / audio_downsample_factor)
+                    wav2vec = self.text_to_semantic.wav2vec
+                    audio_enc_dec = self.voicebox.audio_enc_dec
+
+                    cond_target_length = (cond_tokens_seq_len * wav2vec.target_sample_hz / wav2vec.downsample_factor) / (audio_enc_dec.sampling_rate / audio_enc_dec.downsample_factor)
+                    cond_target_length = math.ceil(cond_target_length)
+
+                elif exists(self.duration_predictor):
+                    cond_target_length = cond_tokens_seq_len
+
+                cond = curtail_or_pad(cond, cond_target_length)
+            else:
+                cond = torch.zeros((cond_token_ids.shape[0], cond_target_length, self.dim_cond_emb), device = self.device)
+        else:
+            assert num_cond_inputs == 0, 'no conditioning inputs should be given if not conditioning on text'
+
+        shape = cond.shape
+        batch = shape[0]
+
+        # neural ode
+
         self.voicebox.eval()
 
         def fn(t, x, *, packed_shape = None):
@@ -880,9 +1160,11 @@ class ConditionalFlowMatcherWrapper(Module):
             out = self.voicebox.forward_with_cond_scale(
                 x,
                 times = t,
-                phoneme_ids = phoneme_ids,
+                cond_token_ids = cond_token_ids,
                 cond = cond,
-                cond_scale = cond_scale
+                cond_scale = cond_scale,
+                cond_mask = cond_mask,
+                self_attn_mask = self_attn_mask
             )
 
             if exists(packed_shape):
@@ -894,12 +1176,12 @@ class ConditionalFlowMatcherWrapper(Module):
         t = torch.linspace(0, 1, steps, device = self.device)
 
         if not self.use_torchode:
-            print('sampling with torchdiffeq')
+            LOGGER.debug('sampling with torchdiffeq')
 
             trajectory = odeint(fn, y0, t, **self.odeint_kwargs)
             sampled = trajectory[-1]
         else:
-            print('sampling with torchode')
+            LOGGER.debug('sampling with torchode')
 
             t = repeat(t, 'n -> b n', b = batch)
             y0, packed_shape = pack_one(y0, 'b *')
@@ -934,9 +1216,12 @@ class ConditionalFlowMatcherWrapper(Module):
         self,
         x1,
         *,
-        phoneme_ids,
-        cond,
-        mask = None
+        mask = None,
+        semantic_token_ids = None,
+        phoneme_ids = None,
+        cond = None,
+        cond_mask = None,
+        input_sampling_rate = None # will assume it to be the same as the audio encoder decoder sampling rate, if not given. if given, will resample
     ):
         """
         following eq (5) (6) in https://arxiv.org/pdf/2306.15687.pdf
@@ -948,18 +1233,50 @@ class ConditionalFlowMatcherWrapper(Module):
 
         input_is_raw_audio, cond_is_raw_audio = map(is_probably_audio_from_shape, (x1, cond))
 
+        if input_is_raw_audio:
+            raw_audio = x1
+
         if any([input_is_raw_audio, cond_is_raw_audio]):
             assert exists(self.voicebox.audio_enc_dec), 'audio_enc_dec must be set on VoiceBox to train directly on raw audio'
 
+            audio_enc_dec_sampling_rate = self.voicebox.audio_enc_dec.sampling_rate
+            input_sampling_rate = default(input_sampling_rate, audio_enc_dec_sampling_rate)
+
             with torch.no_grad():
                 self.voicebox.audio_enc_dec.eval()
 
                 if input_is_raw_audio:
+                    x1 = resample(x1, input_sampling_rate, audio_enc_dec_sampling_rate)
                     x1 = self.voicebox.audio_enc_dec.encode(x1)
 
-                if cond_is_raw_audio:
+                if exists(cond) and cond_is_raw_audio:
+                    cond = resample(cond, input_sampling_rate, audio_enc_dec_sampling_rate)
                     cond = self.voicebox.audio_enc_dec.encode(cond)
 
+        # setup text conditioning, either coming from duration model (as phoneme ids)
+        # or from text-to-semantic module, semantic ids encoded with wav2vec (hubert usually)
+
+        assert self.condition_on_text or not (exists(semantic_token_ids) or exists(phoneme_ids)), 'semantic or phoneme ids should not be passed in if not conditioning on text'
+
+        cond_token_ids = None
+
+        if self.condition_on_text:
+            if exists(self.text_to_semantic) or exists(semantic_token_ids):
+                assert not exists(phoneme_ids), 'phoneme ids are not needed for conditioning with spear-tts text-to-semantic'
+
+                if not exists(semantic_token_ids):
+                    assert input_is_raw_audio
+                    wav2vec = self.text_to_semantic.wav2vec
+                    wav2vec_input = resample(raw_audio, input_sampling_rate, wav2vec.target_sample_hz)
+                    semantic_token_ids = wav2vec(wav2vec_input).clone()
+
+                cond_token_ids = semantic_token_ids
+            else:
+                assert exists(phoneme_ids)
+                cond_token_ids = phoneme_ids
+
+        # main conditional flow logic is below
+
         # x0 is gaussian noise
 
         x0 = torch.randn_like(x1)
@@ -981,11 +1298,12 @@ class ConditionalFlowMatcherWrapper(Module):
 
         loss = self.voicebox(
             w,
-            phoneme_ids = phoneme_ids,
             cond = cond,
-            mask = mask,
+            cond_mask = cond_mask,
             times = times,
             target = flow,
+            self_attn_mask = mask,
+            cond_token_ids = cond_token_ids,
             cond_drop_prob = self.cond_drop_prob
         )
 
